@@ -1,7 +1,6 @@
 /* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+/* kdc/main.c - Main procedure body for the KDC server process */
 /*
- * kdc/main.c
- *
  * Copyright 1990,2001,2008,2009 by the Massachusetts Institute of Technology.
  *
  * Export of this software from the United States of America may
@@ -22,9 +21,6 @@
  * M.I.T. makes no representations about the suitability of
  * this software for any purpose.  It is provided "as is" without express
  * or implied warranty.
- *
- *
- * Main procedure body for the KDC server process.
  */
 /*
  * Copyright (c) 2006-2008, Novell, Inc.
@@ -82,11 +78,6 @@ extern int daemon(int, int);
 
 static void usage (char *);
 
-static krb5_sigtype request_exit (int);
-static krb5_sigtype request_hup  (int);
-
-static void setup_signal_handlers (void);
-
 static krb5_error_code setup_sam (void);
 
 static void initialize_realms (krb5_context, int, char **);
@@ -97,10 +88,8 @@ static int nofork = 0;
 static int workers = 0;
 static const char *pid_file = NULL;
 static int rkey_init_done = 0;
-
-#ifdef POSIX_SIGNALS
-static struct sigaction s_action;
-#endif /* POSIX_SIGNALS */
+static volatile int signal_received = 0;
+static volatile int sighup_received = 0;
 
 #define KRB5_KDC_MAX_REALMS     32
 
@@ -166,6 +155,8 @@ setup_server_realm(krb5_principal sprinc)
 static void
 finish_realm(kdc_realm_t *rdp)
 {
+    if (rdp->realm_name)
+        free(rdp->realm_name);
     if (rdp->realm_mpname)
         free(rdp->realm_mpname);
     if (rdp->realm_stash)
@@ -295,17 +286,21 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
         goto whoops;
     }
 
-    rdp->realm_name = realm;
+    rdp->realm_name = strdup(realm);
+    if (rdp->realm_name == NULL) {
+        kret = ENOMEM;
+        goto whoops;
+    }
     kret = krb5int_init_context_kdc(&rdp->realm_context);
     if (kret) {
-        kdc_err(NULL, kret, "while getting context for realm %s", realm);
+        kdc_err(NULL, kret, _("while getting context for realm %s"), realm);
         goto whoops;
     }
 
     kret = krb5_read_realm_params(rdp->realm_context, rdp->realm_name,
                                   &rparams);
     if (kret) {
-        kdc_err(rdp->realm_context, kret, "while reading realm parameters");
+        kdc_err(rdp->realm_context, kret, _("while reading realm parameters"));
         goto whoops;
     }
 
@@ -396,8 +391,8 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
 
     /* Set the default realm of this context */
     if ((kret = krb5_set_default_realm(rdp->realm_context, realm))) {
-        kdc_err(rdp->realm_context, kret, "while setting default realm to %s",
-                realm);
+        kdc_err(rdp->realm_context, kret,
+                _("while setting default realm to %s"), realm);
         goto whoops;
     }
 
@@ -405,7 +400,7 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
     kdb_open_flags = KRB5_KDB_OPEN_RW | KRB5_KDB_SRV_TYPE_KDC;
     if ((kret = krb5_db_open(rdp->realm_context, db_args, kdb_open_flags))) {
         kdc_err(rdp->realm_context, kret,
-                "while initializing database for realm %s", realm);
+                _("while initializing database for realm %s"), realm);
         goto whoops;
     }
 
@@ -414,7 +409,7 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
                                         rdp->realm_name, (char **) NULL,
                                         &rdp->realm_mprinc))) {
         kdc_err(rdp->realm_context, kret,
-                "while setting up master key name %s for realm %s",
+                _("while setting up master key name %s for realm %s"),
                 rdp->realm_mpname, realm);
         goto whoops;
     }
@@ -427,7 +422,7 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
                                    FALSE, rdp->realm_stash,
                                    &mkvno, NULL, &rdp->realm_mkey))) {
         kdc_err(rdp->realm_context, kret,
-                "while fetching master key %s for realm %s",
+                _("while fetching master key %s for realm %s"),
                 rdp->realm_mpname, realm);
         goto whoops;
     }
@@ -435,7 +430,7 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
     if ((kret = krb5_db_fetch_mkey_list(rdp->realm_context, rdp->realm_mprinc,
                                         &rdp->realm_mkey, mkvno, &rdp->mkey_list))) {
         kdc_err(rdp->realm_context, kret,
-                "while fetching master keys list for realm %s", realm);
+                _("while fetching master keys list for realm %s"), realm);
         goto whoops;
     }
 
@@ -444,7 +439,7 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
     if ((kret = krb5_ktkdb_resolve(rdp->realm_context, NULL,
                                    &rdp->realm_keytab))) {
         kdc_err(rdp->realm_context, kret,
-                "while resolving kdb keytab for realm %s", realm);
+                _("while resolving kdb keytab for realm %s"), realm);
         goto whoops;
     }
 
@@ -453,7 +448,7 @@ init_realm(kdc_realm_t *rdp, char *realm, char *def_mpname,
                                      strlen(realm), realm, KRB5_TGS_NAME,
                                      realm, (char *) NULL))) {
         kdc_err(rdp->realm_context, kret,
-                "while building TGS name for realm %s", realm);
+                _("while building TGS name for realm %s"), realm);
         goto whoops;
     }
 
@@ -485,9 +480,9 @@ whoops:
 }
 
 static krb5_sigtype
-request_exit(int signo)
+on_monitor_signal(int signo)
 {
-    signal_requests_exit = 1;
+    signal_received = signo;
 
 #ifdef POSIX_SIGTYPE
     return;
@@ -497,9 +492,9 @@ request_exit(int signo)
 }
 
 static krb5_sigtype
-request_hup(int signo)
+on_monitor_sighup(int signo)
 {
-    signal_requests_reset = 1;
+    sighup_received = 1;
 
 #ifdef POSIX_SIGTYPE
     return;
@@ -508,45 +503,23 @@ request_hup(int signo)
 #endif
 }
 
-static void
-setup_signal_handlers(void)
-{
-#ifdef POSIX_SIGNALS
-    (void) sigemptyset(&s_action.sa_mask);
-    s_action.sa_flags = 0;
-    s_action.sa_handler = request_exit;
-    (void) sigaction(SIGINT, &s_action, (struct sigaction *) NULL);
-    (void) sigaction(SIGTERM, &s_action, (struct sigaction *) NULL);
-    s_action.sa_handler = request_hup;
-    (void) sigaction(SIGHUP, &s_action, (struct sigaction *) NULL);
-    s_action.sa_handler = SIG_IGN;
-    (void) sigaction(SIGPIPE, &s_action, (struct sigaction *) NULL);
-#else  /* POSIX_SIGNALS */
-    signal(SIGINT, request_exit);
-    signal(SIGTERM, request_exit);
-    signal(SIGHUP, request_hup);
-    signal(SIGPIPE, SIG_IGN);
-#endif /* POSIX_SIGNALS */
-
-    return;
-}
-
 /*
  * Kill the worker subprocesses given by pids[0..bound-1], skipping any which
  * are set to -1, and wait for them to exit (so that we know the ports are no
- * longer in use).  num_active must be the number of active (i.e. not -1) pids
- * in the array.
+ * longer in use).
  */
 static void
-terminate_workers(pid_t *pids, int bound, int num_active)
+terminate_workers(pid_t *pids, int bound)
 {
-    int i, status;
+    int i, status, num_active = 0;
     pid_t pid;
 
     /* Kill the active worker pids. */
     for (i = 0; i < bound; i++) {
-        if (pids[i] != -1)
-            kill(pids[i], SIGTERM);
+        if (pids[i] == -1)
+            continue;
+        kill(pids[i], SIGTERM);
+        num_active++;
     }
 
     /* Wait for them to exit. */
@@ -563,19 +536,59 @@ terminate_workers(pid_t *pids, int bound, int num_active)
  * function in error cases.
  */
 static krb5_error_code
-create_workers(int num)
+create_workers(verto_ctx *ctx, int num)
 {
-    int i, status, numleft;
+    krb5_error_code retval;
+    int i, status;
     pid_t pid, *pids;
+#ifdef POSIX_SIGNALS
+    struct sigaction s_action;
+#endif /* POSIX_SIGNALS */
+
+    /*
+     * Setup our signal handlers which will forward to the children.
+     * These handlers will be overriden in the child processes.
+     */
+#ifdef POSIX_SIGNALS
+    (void) sigemptyset(&s_action.sa_mask);
+    s_action.sa_flags = 0;
+    s_action.sa_handler = on_monitor_signal;
+    (void) sigaction(SIGINT, &s_action, (struct sigaction *) NULL);
+    (void) sigaction(SIGTERM, &s_action, (struct sigaction *) NULL);
+    (void) sigaction(SIGQUIT, &s_action, (struct sigaction *) NULL);
+    s_action.sa_handler = on_monitor_sighup;
+    (void) sigaction(SIGHUP, &s_action, (struct sigaction *) NULL);
+#else  /* POSIX_SIGNALS */
+    signal(SIGINT, on_monitor_signal);
+    signal(SIGTERM, on_monitor_signal);
+    signal(SIGQUIT, on_monitor_signal);
+    signal(SIGHUP, on_monitor_sighup);
+#endif /* POSIX_SIGNALS */
 
     /* Create child worker processes; return in each child. */
-    krb5_klog_syslog(LOG_INFO, "creating %d worker processes", num);
+    krb5_klog_syslog(LOG_INFO, _("creating %d worker processes"), num);
     pids = calloc(num, sizeof(pid_t));
     if (pids == NULL)
         return ENOMEM;
     for (i = 0; i < num; i++) {
         pid = fork();
         if (pid == 0) {
+            if (!verto_reinitialize(ctx)) {
+                krb5_klog_syslog(LOG_ERR,
+                                 _("Unable to reinitialize main loop"));
+                return ENOMEM;
+            }
+            retval = loop_setup_signals(ctx, NULL, reset_for_hangup);
+            if (retval) {
+                krb5_klog_syslog(LOG_ERR, _("Unable to initialize signal "
+                                            "handlers in pid %d"), pid);
+                return retval;
+            }
+
+            /* Avoid race condition */
+            if (signal_received)
+                exit(0);
+
             /* Return control to main() in the new worker process. */
             free(pids);
             return 0;
@@ -583,20 +596,22 @@ create_workers(int num)
         if (pid == -1) {
             /* Couldn't fork enough times. */
             status = errno;
-            terminate_workers(pids, i, i);
+            terminate_workers(pids, i);
             free(pids);
             return status;
         }
         pids[i] = pid;
     }
 
+    /* We're going to use our own main loop here. */
+    loop_free(ctx);
+
     /* Supervise the worker processes. */
-    numleft = num;
-    while (!signal_requests_exit) {
+    while (!signal_received) {
         /* Wait until a worker process exits or we get a signal. */
         pid = wait(&status);
         if (pid >= 0) {
-            krb5_klog_syslog(LOG_ERR, "worker %ld exited with status %d",
+            krb5_klog_syslog(LOG_ERR, _("worker %ld exited with status %d"),
                              (long) pid, status);
 
             /* Remove the pid from the table. */
@@ -611,18 +626,19 @@ create_workers(int num)
         }
 
         /* Propagate HUP signal to worker processes if we received one. */
-        if (signal_requests_reset) {
+        if (sighup_received) {
+            sighup_received = 0;
             for (i = 0; i < num; i++) {
                 if (pids[i] != -1)
                     kill(pids[i], SIGHUP);
             }
-            signal_requests_reset = 0;
         }
     }
-    if (signal_requests_exit)
-        krb5_klog_syslog(LOG_INFO, "shutdown signal received in supervisor");
+    if (signal_received)
+        krb5_klog_syslog(LOG_INFO, _("signal %d received in supervisor"),
+                         signal_received);
 
-    terminate_workers(pids, num, numleft);
+    terminate_workers(pids, num);
     free(pids);
     exit(0);
 }
@@ -637,14 +653,14 @@ static void
 usage(char *name)
 {
     fprintf(stderr,
-            "usage: %s [-x db_args]* [-d dbpathname] [-r dbrealmname]\n"
-            "\t\t[-R replaycachename] [-m] [-k masterenctype]\n"
-            "\t\t[-M masterkeyname] [-p port] [-P pid_file]\n"
-            "\t\t[-n] [-w numworkers] [/]\n\n"
-            "where,\n"
-            "\t[-x db_args]* - Any number of database specific arguments.\n"
-            "\t\t\tLook at each database module documentation for supported\n"
-            "\t\t\targuments\n",
+            _("usage: %s [-x db_args]* [-d dbpathname] [-r dbrealmname]\n"
+              "\t\t[-R replaycachename] [-m] [-k masterenctype]\n"
+              "\t\t[-M masterkeyname] [-p port] [-P pid_file]\n"
+              "\t\t[-n] [-w numworkers] [/]\n\n"
+              "where,\n"
+              "\t[-x db_args]* - Any number of database specific arguments.\n"
+              "\t\t\tLook at each database module documentation for "
+              "\t\t\tsupported arguments\n"),
             name);
     exit(1);
 }
@@ -698,22 +714,20 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                 host_based_srvcs = 0;
         }
 
-        /* aprof_init can return 0 with aprof == NULL */
-        if (aprof)
-            krb5_aprof_finish(aprof);
+        krb5_aprof_finish(aprof);
     }
 
     if (default_udp_ports == 0) {
         default_udp_ports = strdup(DEFAULT_KDC_UDP_PORTLIST);
         if (default_udp_ports == 0) {
-            fprintf(stderr," KDC cannot initialize. Not enough memory\n");
+            fprintf(stderr, _(" KDC cannot initialize. Not enough memory\n"));
             exit(1);
         }
     }
     if (default_tcp_ports == 0) {
         default_tcp_ports = strdup(DEFAULT_KDC_TCP_PORTLIST);
         if (default_tcp_ports == 0) {
-            fprintf(stderr," KDC cannot initialize. Not enough memory\n");
+            fprintf(stderr, _(" KDC cannot initialize. Not enough memory\n"));
             exit(1);
         }
     }
@@ -730,8 +744,8 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                 char **temp = realloc( db_args, sizeof(char*) * (db_args_size+1)); /* one for NULL */
                 if( temp == NULL )
                 {
-                    fprintf(stderr,"%s: KDC cannot initialize. Not enough memory\n",
-                            argv[0]);
+                    fprintf(stderr, _("%s: KDC cannot initialize. Not enough "
+                                      "memory\n"), argv[0]);
                     exit(1);
                 }
 
@@ -749,8 +763,8 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                                              default_tcp_ports, manual,
                                              def_restrict_anon, db_args,
                                              no_refrls, host_based_srvcs))) {
-                        fprintf(stderr,
-                                "%s: cannot initialize realm %s - see log file for details\n",
+                        fprintf(stderr, _("%s: cannot initialize realm %s - "
+                                          "see log file for details\n"),
                                 argv[0], optarg);
                         exit(1);
                     }
@@ -760,8 +774,8 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                 }
                 else
                 {
-                    fprintf(stderr,"%s: cannot initialize realm %s. Not enough memory\n",
-                            argv[0], optarg);
+                    fprintf(stderr, _("%s: cannot initialize realm %s. Not "
+                                      "enough memory\n"), argv[0], optarg);
                     exit(1);
                 }
             }
@@ -772,9 +786,8 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
              */
             if( db_name == NULL ) {
                 if (asprintf(&db_name, "dbname=%s", optarg) < 0) {
-                    fprintf(stderr,
-                            "%s: KDC cannot initialize. Not enough memory\n",
-                            argv[0]);
+                    fprintf(stderr, _("%s: KDC cannot initialize. Not enough "
+                                      "memory\n"), argv[0]);
                     exit(1);
                 }
             }
@@ -784,8 +797,8 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                 char **temp = realloc( db_args, sizeof(char*) * (db_args_size+1)); /* one for NULL */
                 if( temp == NULL )
                 {
-                    fprintf(stderr,"%s: KDC cannot initialize. Not enough memory\n",
-                            argv[0]);
+                    fprintf(stderr, _("%s: KDC cannot initialize. Not enough "
+                                      "memory\n"), argv[0]);
                     exit(1);
                 }
 
@@ -812,7 +825,7 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
             break;
         case 'k':                       /* enctype for master key */
             if (krb5_string_to_enctype(optarg, &menctype))
-                com_err(argv[0], 0, "invalid enctype %s", optarg);
+                com_err(argv[0], 0, _("invalid enctype %s"), optarg);
             break;
         case 'R':
             /* Replay cache name; defunct since we don't use a replay cache. */
@@ -825,7 +838,8 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                 free(default_udp_ports);
             default_udp_ports = strdup(optarg);
             if (!default_udp_ports) {
-                fprintf(stderr," KDC cannot initialize. Not enough memory\n");
+                fprintf(stderr, _(" KDC cannot initialize. Not enough "
+                                  "memory\n"));
                 exit(1);
             }
 #if 0 /* not yet */
@@ -851,8 +865,9 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
         /* no realm specified, use default realm */
         if ((retval = krb5_get_default_realm(kcontext, &lrealm))) {
             com_err(argv[0], retval,
-                    "while attempting to retrieve default realm");
-            fprintf (stderr, "%s: %s, attempting to retrieve default realm\n",
+                    _("while attempting to retrieve default realm"));
+            fprintf (stderr,
+                     _("%s: %s, attempting to retrieve default realm\n"),
                      argv[0], krb5_get_error_message(kcontext, retval));
             exit(1);
         }
@@ -861,13 +876,14 @@ initialize_realms(krb5_context kcontext, int argc, char **argv)
                                      default_udp_ports, default_tcp_ports,
                                      manual, def_restrict_anon, db_args,
                                      no_refrls, host_based_srvcs))) {
-                fprintf(stderr,"%s: cannot initialize realm %s - see log file for details\n",
-                        argv[0], lrealm);
+                fprintf(stderr, _("%s: cannot initialize realm %s - see log "
+                                  "file for details\n"), argv[0], lrealm);
                 exit(1);
             }
             kdc_realmlist[0] = rdatap;
             kdc_numrealms++;
         }
+        krb5_free_default_realm(kcontext, lrealm);
     }
 
     /* Ensure that this is set for our first request. */
@@ -945,15 +961,17 @@ int main(int argc, char **argv)
 {
     krb5_error_code     retval;
     krb5_context        kcontext;
+    verto_ctx *ctx;
     int errout = 0;
     int i;
 
+    setlocale(LC_MESSAGES, "");
     if (strrchr(argv[0], '/'))
         argv[0] = strrchr(argv[0], '/')+1;
 
     if (!(kdc_realmlist = (kdc_realm_t **) malloc(sizeof(kdc_realm_t *) *
                                                   KRB5_KDC_MAX_REALMS))) {
-        fprintf(stderr, "%s: cannot get memory for realm list\n", argv[0]);
+        fprintf(stderr, _("%s: cannot get memory for realm list\n"), argv[0]);
         exit(1);
     }
     memset(kdc_realmlist, 0,
@@ -967,7 +985,7 @@ int main(int argc, char **argv)
      */
     retval = krb5int_init_context_kdc(&kcontext);
     if (retval) {
-        com_err(argv[0], retval, "while initializing krb5");
+        com_err(argv[0], retval, _("while initializing krb5"));
         exit(1);
     }
     krb5_klog_init(kcontext, "kdc", argv[0], 1);
@@ -985,14 +1003,19 @@ int main(int argc, char **argv)
      */
     initialize_realms(kcontext, argc, argv);
 
-    setup_signal_handlers();
+    ctx = loop_init(VERTO_EV_TYPE_NONE);
+    if (!ctx) {
+        kdc_err(kcontext, ENOMEM, _("while creating main loop"));
+        finish_realms();
+        return 1;
+    }
 
     load_preauth_plugins(kcontext);
     load_authdata_plugins(kcontext);
 
     retval = setup_sam();
     if (retval) {
-        kdc_err(kcontext, retval, "while initializing SAM");
+        kdc_err(kcontext, retval, _("while initializing SAM"));
         finish_realms();
         return 1;
     }
@@ -1009,7 +1032,7 @@ int main(int argc, char **argv)
             port = strtol(cp, &cp, 10);
             if (cp == 0)
                 break;
-            retval = add_udp_port(port);
+            retval = loop_add_udp_port(port);
             if (retval)
                 goto net_init_error;
         }
@@ -1023,7 +1046,7 @@ int main(int argc, char **argv)
             port = strtol(cp, &cp, 10);
             if (cp == 0)
                 break;
-            retval = add_tcp_port(port);
+            retval = loop_add_tcp_port(port);
             if (retval)
                 goto net_init_error;
         }
@@ -1035,44 +1058,56 @@ int main(int argc, char **argv)
      * children won't be able to re-open the listener sockets.  Hopefully our
      * platform has pktinfo support and doesn't need reconfigs.
      */
-    if ((retval = setup_network(NULL, kdc_progname, (workers > 0)))) {
+    if (workers == 0) {
+        retval = loop_setup_routing_socket(ctx, NULL, kdc_progname);
+        if (retval) {
+            kdc_err(kcontext, retval, _("while initializing routing socket"));
+            finish_realms();
+            return 1;
+        }
+        retval = loop_setup_signals(ctx, NULL, reset_for_hangup);
+        if (retval) {
+            kdc_err(kcontext, retval, _("while initializing signal handlers"));
+            finish_realms();
+            return 1;
+        }
+    }
+    if ((retval = loop_setup_network(ctx, NULL, kdc_progname))) {
     net_init_error:
-        kdc_err(kcontext, retval, "while initializing network");
+        kdc_err(kcontext, retval, _("while initializing network"));
         finish_realms();
         return 1;
     }
     if (!nofork && daemon(0, 0)) {
-        kdc_err(kcontext, errno, "while detaching from tty");
+        kdc_err(kcontext, errno, _("while detaching from tty"));
         finish_realms();
         return 1;
     }
     if (pid_file != NULL) {
         retval = write_pid_file(pid_file);
         if (retval) {
-            kdc_err(kcontext, retval, "while creating PID file");
+            kdc_err(kcontext, retval, _("while creating PID file"));
             finish_realms();
             return 1;
         }
     }
     if (workers > 0) {
         finish_realms();
-        retval = create_workers(workers);
+        retval = create_workers(ctx, workers);
         if (retval) {
-            kdc_err(kcontext, errno, "creating worker processes");
+            kdc_err(kcontext, errno, _("creating worker processes"));
             return 1;
         }
         /* We get here only in a worker child process; re-initialize realms. */
         initialize_realms(kcontext, argc, argv);
     }
-    krb5_klog_syslog(LOG_INFO, "commencing operation");
+    krb5_klog_syslog(LOG_INFO, _("commencing operation"));
     if (nofork)
-        fprintf(stderr, "%s: starting...\n", kdc_progname);
-    if ((retval = listen_and_process(0, kdc_progname, reset_for_hangup))) {
-        kdc_err(kcontext, retval, "while processing network requests");
-        errout++;
-    }
-    closedown_network();
-    krb5_klog_syslog(LOG_INFO, "shutting down");
+        fprintf(stderr, _("%s: starting...\n"), kdc_progname);
+
+    verto_run(ctx);
+    loop_free(ctx);
+    krb5_klog_syslog(LOG_INFO, _("shutting down"));
     unload_preauth_plugins(kcontext);
     unload_authdata_plugins(kcontext);
     krb5_klog_close(kdc_context);
