@@ -371,11 +371,11 @@ static struct mech_attr_info_desc {
 
 OM_uint32
 generic_gss_display_mech_attr(
-      OM_uint32         *minor_status,
-      gss_const_OID      mech_attr,
-      gss_buffer_t       name,
-      gss_buffer_t       short_desc,
-      gss_buffer_t       long_desc)
+    OM_uint32         *minor_status,
+    gss_const_OID      mech_attr,
+    gss_buffer_t       name,
+    gss_buffer_t       short_desc,
+    gss_buffer_t       long_desc)
 {
     size_t i;
 
@@ -412,3 +412,10 @@ generic_gss_display_mech_attr(
 
     return GSS_S_BAD_MECH_ATTR;
 }
+
+static gss_buffer_desc const_attrs[] = {
+    { sizeof("local-login-user") - 1,
+      "local-login-user" },
+};
+
+GSS_DLLIMP gss_buffer_t GSS_C_ATTR_LOCAL_LOGIN_USER = &const_attrs[0];
