@@ -1,7 +1,6 @@
 /* -*- mode: c; c-basic-offset: 4; indent-tabs-mode: nil -*- */
+/* lib/gssapi/generic/oid_ops.c */
 /*
- * lib/gssapi/generic/oid_ops.c
- *
  * Copyright 1995 by the Massachusetts Institute of Technology.
  * All Rights Reserved.
  *
@@ -23,12 +22,30 @@
  * M.I.T. makes no representations about the suitability of
  * this software for any purpose.  It is provided "as is" without express
  * or implied warranty.
+ */
+/*
+ * Copyright 1993 by OpenVision Technologies, Inc.
  *
+ * Permission to use, copy, modify, distribute, and sell this software
+ * and its documentation for any purpose is hereby granted without fee,
+ * provided that the above copyright notice appears in all copies and
+ * that both that copyright notice and this permission notice appear in
+ * supporting documentation, and that the name of OpenVision not be used
+ * in advertising or publicity pertaining to distribution of the software
+ * without specific, written prior permission. OpenVision makes no
+ * representations about the suitability of this software for any
+ * purpose.  It is provided "as is" without express or implied warranty.
+ *
+ * OPENVISION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
+ * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
+ * EVENT SHALL OPENVISION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
+ * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
+ * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
+ * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
+ * PERFORMANCE OF THIS SOFTWARE.
  */
 
-/*
- * oid_ops.c - GSS-API V2 interfaces to manipulate OIDs
- */
+/* GSS-API V2 interfaces to manipulate OIDs */
 
 #include "gssapiP_generic.h"
 #ifdef HAVE_UNISTD_H
@@ -112,7 +129,7 @@ generic_gss_create_empty_oid_set(OM_uint32 *minor_status, gss_OID_set *oid_set)
 {
     *minor_status = 0;
 
-    if ((*oid_set = (gss_OID_set) malloc(sizeof(gss_OID_set_desc)))) {
+    if ((*oid_set = (gss_OID_set) gssalloc_malloc(sizeof(gss_OID_set_desc)))) {
         memset(*oid_set, 0, sizeof(gss_OID_set_desc));
         return(GSS_S_COMPLETE);
     }
@@ -138,8 +155,8 @@ generic_gss_add_oid_set_member(OM_uint32 *minor_status,
 
     elist = (*oid_set)->elements;
     /* Get an enlarged copy of the array */
-    if (((*oid_set)->elements = (gss_OID) malloc(((*oid_set)->count+1) *
-                                                 sizeof(gss_OID_desc)))) {
+    if (((*oid_set)->elements = (gss_OID) gssalloc_malloc(((*oid_set)->count+1) *
+                                                          sizeof(gss_OID_desc)))) {
         /* Copy in the old junk */
         if (elist)
             memcpy((*oid_set)->elements,
@@ -149,7 +166,7 @@ generic_gss_add_oid_set_member(OM_uint32 *minor_status,
         /* Duplicate the input element */
         lastel = &(*oid_set)->elements[(*oid_set)->count];
         if ((lastel->elements =
-             (void *) malloc((size_t) member_oid->length))) {
+             (void *) gssalloc_malloc((size_t) member_oid->length))) {
             /* Success - copy elements */
             memcpy(lastel->elements, member_oid->elements,
                    (size_t) member_oid->length);
@@ -159,12 +176,12 @@ generic_gss_add_oid_set_member(OM_uint32 *minor_status,
             /* Update count */
             (*oid_set)->count++;
             if (elist)
-                free(elist);
+                gssalloc_free(elist);
             *minor_status = 0;
             return(GSS_S_COMPLETE);
         }
         else
-            free((*oid_set)->elements);
+            gssalloc_free((*oid_set)->elements);
     }
     /* Failure - restore old contents of list */
     (*oid_set)->elements = elist;
@@ -253,9 +270,7 @@ generic_gss_oid_to_str(OM_uint32 *minor_status,
         *minor_status = ENOMEM;
         return(GSS_S_FAILURE);
     }
-    oid_str->length = krb5int_buf_len(&buf)+1;
-    oid_str->value = (void *) bp;
-    return(GSS_S_COMPLETE);
+    return k5buf_to_gss(minor_status, &buf, oid_str);
 }
 
 OM_uint32
@@ -478,27 +493,6 @@ generic_gss_oid_decompose(OM_uint32 *minor_status,
     return GSS_S_COMPLETE;
 }
 
-/*
- * Copyright 1993 by OpenVision Technologies, Inc.
- *
- * Permission to use, copy, modify, distribute, and sell this software
- * and its documentation for any purpose is hereby granted without fee,
- * provided that the above copyright notice appears in all copies and
- * that both that copyright notice and this permission notice appear in
- * supporting documentation, and that the name of OpenVision not be used
- * in advertising or publicity pertaining to distribution of the software
- * without specific, written prior permission. OpenVision makes no
- * representations about the suitability of this software for any
- * purpose.  It is provided "as is" without express or implied warranty.
- *
- * OPENVISION DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
- * INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
- * EVENT SHALL OPENVISION BE LIABLE FOR ANY SPECIAL, INDIRECT OR
- * CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
- * USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
- * OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
- * PERFORMANCE OF THIS SOFTWARE.
- */
 OM_uint32
 generic_gss_copy_oid_set(OM_uint32 *minor_status,
                          const gss_OID_set_desc * const oidset,
@@ -521,13 +515,13 @@ generic_gss_copy_oid_set(OM_uint32 *minor_status,
     if (new_oidset == NULL)
         return (GSS_S_CALL_INACCESSIBLE_WRITE);
 
-    if ((copy = (gss_OID_set_desc *) calloc(1, sizeof (*copy))) == NULL) {
+    if ((copy = (gss_OID_set_desc *) gssalloc_calloc(1, sizeof (*copy))) == NULL) {
         major = GSS_S_FAILURE;
         goto done;
     }
 
     if ((copy->elements = (gss_OID_desc *)
-         calloc(oidset->count, sizeof (*copy->elements))) == NULL) {
+         gssalloc_calloc(oidset->count, sizeof (*copy->elements))) == NULL) {
         major = GSS_S_FAILURE;
         goto done;
     }
@@ -537,7 +531,7 @@ generic_gss_copy_oid_set(OM_uint32 *minor_status,
         gss_OID_desc *out = &copy->elements[i];
         gss_OID_desc *in = &oidset->elements[i];
 
-        if ((out->elements = (void *) malloc(in->length)) == NULL) {
+        if ((out->elements = (void *) gssalloc_malloc(in->length)) == NULL) {
             major = GSS_S_FAILURE;
             goto done;
         }
