@@ -133,7 +133,7 @@ unsigned int
 g_token_size(const gss_OID_desc * mech, unsigned int body_size)
 {
     /* set body_size to sequence contents size */
-    body_size += 4 + (int) mech->length;         /* NEED overflow check */
+    body_size += 4 + (unsigned int)mech->length;         /* NEED overflow check */
     return(1 + der_length_size(body_size) + body_size);
 }
 
@@ -148,7 +148,7 @@ g_make_token_header(
     int tok_type)
 {
     *(*buf)++ = 0x60;
-    der_write_length(buf, (tok_type == -1) ?2:4 + mech->length + body_size);
+    der_write_length(buf, ((tok_type == -1) ? 2 : 4) + mech->length + body_size);
     *(*buf)++ = 0x06;
     *(*buf)++ = (unsigned char) mech->length;
     TWRITE_STR(*buf, mech->elements, mech->length);
